#!/bin/env python
#
# A simple library to boot a new Linux kernel with kexec
#
# Copyright (C) 2010 Marko Myllynen <myllynen@redhat.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import subprocess

class Kexec:
    """Execute a new Linux kernel with kexec"""

    def __init__(self, kernel, initrd=None, cmdline=None):
        self.prepared = False
        self.kernel = kernel
        self.initrd = initrd
        self.cmdline = cmdline
        if not self.cmdline:
            self.cmdline = file('/proc/cmdline').read()[:-1]

        f = open('/sbin/kexec')
        f.close()
        f = open(self.kernel)
        f.close()
        if self.initrd:
            f = open(self.initrd)
            f.close()

    def prepare(self, message=None):
        if self.initrd:
            r = subprocess.call(['/sbin/kexec',
                                    '--load', self.kernel,
                                    '--append=%s' % self.cmdline,
                                    '--initrd=%s' % self.initrd])
        else:
            r = subprocess.call(['/sbin/kexec',
                                    '--load', self.kernel,
                                    '--append=%s' % self.cmdline])

        if r != 0:
            raise Exception("Failed to load the kernel (kexec returned %d)" % r)

        if not message:
            message = "The system is going down for reboot NOW!!!"

        r = subprocess.call(['/bin/echo', message])
        r = subprocess.call(['/usr/bin/wall', message])

        if r != 0:
            raise Exception("Failed to prepare the system (wall returned %d)" % r)

        self.prepared = True

    def reboot(self):
        if not self.prepared:
            raise Exception("The system has not been prepared for reboot yet")

        r = subprocess.call(['/usr/bin/reboot'])

        if r != 0:
            raise Exception("Failed to reboot the system (reboot returned %d)" % r)

    def unprepare(self, message=None):
        r = subprocess.call(['/sbin/kexec', '--unload'])

        if r != 0:
            raise Exception("Failed to unprepare the system (kexec returned %d)" % r)

        if not message:
            message = "Shutdown cancelled."

        r = subprocess.call(['/bin/echo', message])
        r = subprocess.call(['/usr/bin/wall', message])

        self.prepared = False

    def forceboot(self):
        if not self.prepared:
            raise Exception("The system has not been prepared for reboot yet")

        r = subprocess.call(['/usr/bin/chvt', '1'])
        if r != 0:
            sys.stderr.write("Warning: failed to change virtual terminal.\n")

        r = subprocess.call(['/sbin/kexec', '--exec'])

        if r != 0:
            raise Exception("Failed to reboot the system (kexec returned %d)" % r)

def main():
    kex = Kexec(sys.argv[1], sys.argv[2])
    kex.prepare()
    kex.reboot()

if __name__ == "__main__":
    main()
