#!/bin/env python
#
# A simple library to use PXE
#
# Copyright (C) 2010 Marko Myllynen <myllynen@redhat.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys
import socket
import struct
import subprocess
import tempfile

try:
    from snack import *
except:
    pass

import kexec

class PXE:
    """Runtime PXE"""

    def __init__(self):
        self.base_url = None
        self.main_menu_url = None

    def get_pxe_mac_hex(self, interface=None):
        if not interface:
            cmd = "/sbin/ifconfig"
        else:
            cmd = "/sbin/ifconfig " + interface
        mac = None
        hex = None
        output = os.popen(cmd).readlines()

        for line in output:
            if line.find("Ether") > -1:
                mac = line.split()[4]
                hex = None
            if mac and line.find("inet addr:") > -1:
                hex = line.split()[1]
                hex = hex.replace('addr:', '')
                hex = struct.unpack("!I", socket.inet_aton(hex))
                break

        if hex == None:
            return None, None

        return mac.replace(':', '-').lower(), "%X" % hex

    def get_config_file(self, url, interface=None, screen=None):
        """Download a PXE configuration"""
        configs = list()

        # Internally used for direct downloads if no trailing slash
        if not self.base_url:
            if not re.search("/$", url):
                # First run is never a direct download
               	url = url + "/"
            self.base_url = url

        if re.search("/$", url):
            (mac, hex) = self.get_pxe_mac_hex(interface)
            # As per the PXE specification
            if mac and hex:
                configs.append(url + "pxelinux.cfg/" + "01-" + mac)
                configs.append(url + "pxelinux.cfg/" + hex)
                for i in range(1,8):
                    configs.append(url + "pxelinux.cfg/" + hex[:-i])
            configs.append(url + "pxelinux.cfg/" + "default")
        else:
            # Direct download of a new configuration
            if not re.match("http", url) and not re.match("ftp", url):
                url = self.base_url + url
            configs.append(url)

        if screen:
            grid = GridForm(screen, "Downloading", 1, 3)
            progress = Scale(65, 100)
            grid.add(progress, 0, 1, (0, 1, 0, 0))
            info = Textbox(65, 2, "", wrap = 1)
            grid.add(info, 0, 2, (0, 1, 0, 0))
            grid.draw()
            screen.refresh()

        f = tempfile.NamedTemporaryFile(prefix='pxe-conf.')
        conffile = f.name
        f.close()

        for index, conf in enumerate(configs):
            if screen:
                progress.set(int(100/len(configs)*index))
                conf_file = conf.replace(self.base_url, '')
                info.setText("Trying to download %s..." % conf_file)
                grid.draw()
                screen.refresh()

            r = subprocess.call(['/usr/bin/wget', '-q', conf, '-O', conffile])
            if r == 0:
                if not self.main_menu_url:
                    self.main_menu_url = conf
                if screen:
                    progress.set(100)
                    grid.draw()
                    screen.refresh()
                    screen.popWindow()
                break

        if not os.stat(conffile).st_size:
            self.base_url = None
            raise IOError("File not found")
        return conffile

    def parse_config(self, config_file):
        """Parse a PXE configuration file"""
        f = open(config_file)
        conf = f.readlines(1024)
        f.close()
        menu = PXEMenu()
        label = url = menu_label = kernel = initrd = append = bootif = help = None
        read_main_help = read_entry_help = False
        for line in conf:
            line = line.strip()

            # Help texts are in multiple lines
            if read_main_help:
                if re.match("endtext", line, re.IGNORECASE):
                    read_main_help = False
                    continue
                if menu.help:
                    menu.help = menu.help + " " + line
                else:
                    menu.help = line
                continue
            if read_entry_help:
                if re.match("endtext", line, re.IGNORECASE):
                    read_entry_help = False
                    continue
                if help:
                    help = help + " " + line
                else:
                    help = line
                continue

            # Ignore localboot target
            if re.match("localboot", line, re.IGNORECASE):
                label = url = menu_label = kernel = initrd = append = bootif = help = None

            # A new target
            elif re.match("label", line, re.IGNORECASE):
                if label:
                    # A new label, store the previous one and reset all values
                    if not menu_label:
                        menu_label = label
                    menu.entries.append(PXEMenuEntry(label, url, menu_label,
                                                     kernel, initrd, append,
                                                     bootif, help))
                    label = url = menu_label = kernel = initrd = append = bootif = help = None
                label = ' '.join(line.split()[1:])

            # Handle sub menu as a target
            elif re.match("menu include", line, re.IGNORECASE):
                # XXX handle non-submenu includes
                if label:
                    # A new label, store the previous one and reset all values
                    if not menu_label:
                        menu_label = label
                    if url:
                        # Pretty print to indicate a sub menu
                        menu_label = menu_label + " >"
                    menu.entries.append(PXEMenuEntry(label, url, menu_label,
                                                     kernel, initrd, append,
                                                     bootif, help))
                    label = url = menu_label = kernel = initrd = append = bootif = help = None
                entry = line.split()[2:]
                label = ' '.join(entry[1:])
                url = entry[0]

            # Main level, not under a label yet
            elif not label:
                if re.match("default", line, re.IGNORECASE):
                    menu.default = ' '.join(line.split()[1:])
                elif re.match("prompt", line, re.IGNORECASE):
                    menu.prompt = ' '.join(line.split()[1:])
                elif re.match("menu title", line, re.IGNORECASE):
                    menu.title = ' '.join(line.split()[2:])
                elif re.match("text help", line, re.IGNORECASE):
                    read_main_help = True
                elif re.match("endtext", line, re.IGNORECASE):
                    read_main_help = False
                elif re.match("menu background", line, re.IGNORECASE):
                    menu.background = ' '.join(line.split()[2:])

            # Parsing under a target
            elif re.match("menu label", line, re.IGNORECASE):
                menu_label = ' '.join(line.split()[2:])
                if menu_label == "":
                    # Ignore spacer entries
                    label = None
            elif re.match("kernel", line, re.IGNORECASE):
                kernel = line.split()[1]
            elif re.match("initrd", line, re.IGNORECASE):
                initrd = line.split()[1]
            elif re.match("append", line, re.IGNORECASE):
                append = ' '.join(line.split()[1:])
                if append == "-":
                    append = None
            elif re.match("text help", line, re.IGNORECASE):
                read_entry_help = True
            elif re.match("endtext", line, re.IGNORECASE):
                read_entry_help = False
            elif re.match("ipappend", line, re.IGNORECASE):
                val = line.split()[1]
                if val == 1 or val == 3:
                    # XXX handle as specified
                    val = val
                if val == 2 or val == 3:
                    (mac, hex) = self.get_pxe_mac_hex()
                    bootif = mac
            elif re.match("menu exit", line, re.IGNORECASE):
                url = self.main_menu_url

        # Store the last label
        if label:
            if not menu_label:
                menu_label = label
            if url:
                menu_label = menu_label + " >"
            menu.entries.append(PXEMenuEntry(label, url, menu_label, kernel,
                                             initrd, append, bootif, help))

        return menu

    def get_entry_data(self, menu, label):
        """Get data for the label"""
        for entry in menu.entries:
            if entry.label == label:
                return entry.url, entry.kernel, entry.initrd, entry.append
        return None, None, None, None

    def display_menu(self, menu, screen=None):
        """Display a PXE menu"""

        if not screen:
            if menu.title:
                print menu.title
            if menu.help:
                print menu.help
            for entry in menu.entries:
                print "label: " + entry.label,
                if entry.help:
                    print "(" + entry.help + ")",
                print
            print "\nEnter label to boot:"
            label = sys.stdin.readline().strip()
            for entry in menu.entries:
                if entry.label == label:
                    return entry.url, entry.kernel, entry.initrd, entry.append

        if screen:
            entries = list()
            fullhelp = ""
            for index, entry in enumerate(menu.entries):
                entries.append(entry.menu_label)
                if entry.help:
                    fullhelp = fullhelp + entry.menu_label + ":\n" + entry.help + "\n"
            title = menu.title
            if not title:
                title = "PXE Menu"
            help = menu.help
            if not help:
                help = "Please select a target or sub menu:"
            helpFunc = None
            if fullhelp != "":
                def helpFunc(screen, text):
                    ButtonChoiceWindow(screen, "Help", text, buttons = ['Ok'])
                screen.helpCallback(helpFunc)
            reply = ListboxChoiceWindow(screen, title, help, entries,
                                        help=fullhelp)
            if fullhelp != "":
                screen.helpCallback(None)

            if reply[0] == 'cancel':
                return None, None, None, None

            choice = reply[1]
            return (menu.entries[choice].url, menu.entries[choice].kernel,
                    menu.entries[choice].initrd, menu.entries[choice].append)

        return None, None, None, None

    def download(self, kernel, initrd=None, screen=None):
        """Download a PXE target"""
        if screen:
            grid = GridForm(screen, "Downloading", 1, 4)
            info = Textbox(72, 4, "", wrap=1)
            grid.add(info, 0, 0)
            grid.draw()
            screen.refresh()
            info.setText("Downloading kernel:\n\n%s" % kernel)
            grid.draw()
            screen.refresh()

        f = tempfile.NamedTemporaryFile(prefix='pxe-kernel.')
        kernfile = f.name
        f.close()
        r = subprocess.call(['/usr/bin/wget', '-q', kernel, '-O', kernfile])
        if r != 0:
            if screen:
                screen.popWindow()
            raise Exception("Failed to download the kernel (wget returned %d)" % r)

        initrdfile = None
        if initrd:
            if screen:
                info.setText("Downloading initrd:\n\n%s" % initrd)
                grid.draw()
                screen.refresh()
            f = tempfile.NamedTemporaryFile(prefix='pxe-initrd.')
            initrdfile = f.name
            f.close()
            r = subprocess.call(['/usr/bin/wget', '-q', initrd, '-O', initrdfile])
            if r != 0:
                if screen:
                    screen.popWindow()
                raise Exception("Failed to download the initrd (wget returned %d)" % r)

        return kernfile, initrdfile

class PXEMenu:
    """PXE Menu"""
    def __init__(self, entries=None):
        self.background = None
        self.default = None
        self.prompt = False
        self.title = None
        self.help = None
        self.entries = entries
        if not self.entries:
            self.entries = list()

class PXEMenuEntry:
    """PXE Menu entry"""
    def __init__(self, label, url=None, menu_label=None, kernel=None,
                 initrd=None, append=None, bootif=None, help=None):
        self.label = label
        self.url = url
        self.menu_label = menu_label
        self.kernel = kernel
        self.initrd = initrd
        self.append = append
        self.bootif = bootif
        self.help = help

def main():
    pxe = PXE()
    url = sys.argv[1]
    kernel = None
    while not kernel:
        if not url:
            break
        conf = pxe.get_config(url)
        menu = pxe.parse_config(conf)
        (url, kernel, initrd, append) = pxe.display_menu(menu)

    if not kernel:
        sys.exit(2)

    (kernel_path, initrd_path) = pxe.download(kernel, initrd)

    kex = kexec.Kexec(kernel_path, initrd_path, append)
    kex.prepare("Rebooting to kernel " + kernel_path)
    kex.reboot()

if __name__ == "__main__":
    main()
